var matchdep = require('matchdep');

module.exports = function(grunt) {

  // load all grunt plugins from node_modules folder
  matchdep.filterAll('grunt-*').forEach(grunt.loadNpmTasks);

  var concat = {
      qunit:{
        src:['dist/nucleo-amd.browser.js', 'tests/integration/main.js', 'tests/**/*.spec.js'],
        dest:'tests/setup/qunit/tests.js'
      },
      modules:{
        options: {
          process: function(src, filepath) {

            var splited = filepath.split('/');
            var file_name = splited[splited.length -1].replace('.js', '');

            return 'var ' + file_name + ' = ' + src + ';';

          }
        },
        src: 'lib/modules/**/*.js',
        dest:'dist/modules.js'
      },
      helpers:{
        options: {
          process: function(src, filepath) {

            var splited = filepath.split('/');
            var file_name = splited[splited.length -1].replace('.js', '');

            return 'helpers.' + file_name + ' = ' + src + ';';

          }
        },
        src: 'lib/_helpers/**/*.js',
        dest:'dist/helpers.js'
      },
      browser: {
        options: {
          banner: '(function(mainContext) {\n',
          footer: '\n mainContext.nucleoAMD = NMD; })(this);'
        },
        src: ['lib/main.js'],
        dest:'dist/nucleo-amd.browser.js'
      },
      nodejs: {
        options: {
          footer: '\n module.exports = NMD;'
        },
        src: ['lib/main.js'],
        dest:'dist/nucleo-amd.nodejs.js'
      },
      binaries: {
        files:{
          'bin/nucleo-amd.nodejs.js': ['dist/nucleo-amd.nodejs.js'],
          'bin/nucleo-amd.browser.js':['dist/nucleo-amd.browser.js']
        }
      }
    };



  grunt.initConfig({
    pkg: grunt.file.readJSON('package.json'),
    concat: concat,
    uglify: {
        options: {
            mangle:     true,
            compress:   true,
            banner:     '/* <%= pkg.name %> v<%= pkg.version %>  */'
        },
        prod:    {
            files: {
                'nucleo-amd.min.js': ['nucleo-amd.browser.js']
            }
        }
    },
    express: {
        dev: {
            options: {
                script: 'tests/setup/server/server.js'
            }
        }
    },
    shell: {
        pre_commit_tests: {
          command: 'testem ci'
        }
    },
    watch: {
        options: {
            livereload: true,
            nospawn:    true
        },
        concat: {
            files: [
                'lib/**/*.js',
                'tests/integration/**/*.js'
            ],
            tasks: ['concat:modules', 'concat:helpers', 'concat:browser', 'concat:qunit']
        },
    },
    bump: {
      options: {
        files: ['package.json', 'bower.json'],
        updateConfigs: ['pkg'],
        commit: true,
        commitMessage: 'Release v%VERSION%',
        commitFiles: ['-a'],
        createTag: true,
        tagName: '%VERSION%',
        tagMessage: '%VERSION%',
        push: true,
        pushTo: 'origin',
        gitDescribeOptions: '--tags --always --abbrev=1 --dirty=-d',
        globalReplace: true,
        prereleaseName: false,
        regExp: false
      }
    }

  });

// TODOS:
// - include a task to automaticaly create changelogs based on commits - https://github.com/lalitkapoor/github-changes

  grunt.registerTask('testing', ['concat:browser', 'concat:qunit', 'express', 'watch']);

  grunt.registerTask('build:lib', ['concat:nodejs', 'concat:browser']);
  grunt.registerTask('build:prod', ['build:lib']);

  grunt.registerTask('deploy:commit', ['build:prod', 'concat:binaries', 'bump-commit']);
  grunt.registerTask('deploy:tests', ['shell:pre_commit_tests', 'bump-only:patch', 'uglify']);
  grunt.registerTask('deploy:patch', ['shell:pre_commit_tests', 'bump-only:patch', 'deploy:commit']);
  grunt.registerTask('deploy:minor', ['shell:pre_commit_tests', 'bump-only:minor', 'deploy:commit']);
  grunt.registerTask('deploy:major', ['shell:pre_commit_tests', 'bump-only:major', 'deploy:commit']);
  grunt.registerTask('deploy',       ['deploy:patch']);

  grunt.registerTask('default', ['build:prod']);

};