var NMD = function(options){

    var Mocked = options && options.isMocked ? null : NMD({isMocked:true});
    var registry = {};
    var seen = {};
    var _config = {};
    var defaultConfig = {
        relativePaths: true,
        defaultFile: 'main',
        paths: null
    };

    var helpers = {};

    helpers.cloneObject = function(source) {

        var name, s, empty = {}, dest = {};
        var isNotObject = !helpers.isObject(source);

        if(isNotObject){
            return source;
        }

        for(name in source){
            /*jshint -W089 */
            // the (!(name in empty) || empty[name] !== s) condition avoids copying properties in "source"
            // inherited from Object.prototype.  For example, if dest has a custom toString() method,
            // don't overwrite it with the toString() method that source inherited from Object.prototype
            s = source[name];
            if(!(name in dest) || (dest[name] !== s && (!(name in empty) || empty[name] !== s))){
                dest[name] = helpers.cloneObject ? helpers.cloneObject(s) : s;
            }
        }

        return dest;

    };
    helpers.extend = function(obj1, obj2) {

        var obj3 = {};
        var isObject = helpers.isObject;
        var cloneObject = helpers.cloneObject;
        var extend = helpers.extend;

        if (isObject(obj1)) {

            for (var attrName in obj1) {

                obj3[attrName] = cloneObject(obj1[attrName]);

            }

        }

        if (isObject(obj2)) {

            for (var attrname in obj2) {

                if (isObject(obj1) && isObject(obj1[attrname]) && isObject(obj2[attrname])) {

                    obj3[attrname] = cloneObject(extend(obj1[attrname], obj2[attrname]));

                } else {

                    obj3[attrname] = obj2[attrname];

                }

            }

        }

        return cloneObject(obj3);

    };
    helpers.getDependenciesValues = function(exports, name) {

        var mod = registry[name];
        var deps = mod.deps;
        var reified = [];
        var length = deps.length;
        var dep_name, dep;

        for (var i = 0; i < length; i++) {

            if (deps[i] === 'exports') {

                reified.push(exports);

            } else {

                dep_name = helpers.resolve(deps[i], name);
                dep = helpers.internalRequire(dep_name, name);

                reified.push(dep);

            }

        }

        return reified;

    };
    helpers.getIgnoredPatterns = function() {

        var prefixExitsOnConfig = _config && _config.paths && typeof _config.paths.prefix === 'string';
        var hasIgnore = prefixExitsOnConfig && _config.paths.ignore;
        var ignoreIsString = hasIgnore && typeof _config.paths.ignore === 'string';
        var ignorePatterns = [];

        if (hasIgnore) {

            if (ignoreIsString) {

                ignorePatterns.push(_config.paths.ignore);

            } else if (helpers.isArray(_config.paths.ignore)) {

                ignorePatterns = _config.paths.ignore;

            }

        }

        return ignorePatterns;

    };
    helpers.getMockedModule = function(name) {

        var hasMocks = Mocked && Mocked.has_entries();
        var mocked_module;

        if (!hasMocks) {
            return mocked_module;
        }

        try {

            mocked_module = Mocked.require(name);

        } catch (err) {}


        if (mocked_module !== undefined) {

            // TODO make tests for non-default exported in the module. don't assume that module is 'default' 
            var result = {
                'default': mocked_module
            };

            return result;

        }

        return undefined;

    };
    helpers.getMockedDependencies = function(exports, name) {

        var hasMocks = Mocked && Mocked.has_entries();

        if(!hasMocks){
            return [];
        }

        var reified = helpers.getDependenciesValues(exports, name);

        return reified;

    };
    helpers.getNameWithPrefix = function(path) {

        var prefixExitsOnConfig = _config && _config.paths && typeof _config.paths.prefix === 'string';
        var prefix = prefixExitsOnConfig ? _config.paths.prefix : '';
        var ignorePatterns = helpers.getIgnoredPatterns();
        var shouldIgnore = helpers.shouldIgnorePrefix(path, ignorePatterns);
        var shouldAddPrefix = prefixExitsOnConfig && !shouldIgnore;

        prefix = prefix.length && prefix.slice(-1) !== '/' ? prefix + '/' : prefix;

        var name = shouldAddPrefix ? prefix + path : path;

        return name;

    };
    helpers.getUID = function(options){

        var s4 = function() {
            return Math.floor((1 + Math.random()) * 0x10000)
                .toString(16)
                .substring(1);
        };

        var appended = options && options.append ? options.append : '';
        var uid = s4() + s4() + '-' + s4() + '-' + s4() + '-' + s4() + '-' + s4() + s4() + s4();

        if(appended){
            uid += '/' + appended;
        }

        return uid;

    };
    helpers.handleDefaultFilename = function(name, referrerName){

        if (!registry[name]) {

            var hasEndingSlash = name.slice(-1) === '/';
            var slash = hasEndingSlash ? '' : '/';
            var indexFileName = name + slash + _config.defaultFile;

            if (registry[indexFileName]) {

                name = indexFileName;

            } else if (referrerName) {

                helpers.throwError('Could not find module ' + name + ' required by: ' + referrerName);

            } else {

                helpers.throwError('Could not find module ' + name);

            }

        }

        return name;

    };
    helpers.internalRequire = function(name, referrerName) {

        name = helpers.getNameWithPrefix(name);

        var mocked_module = helpers.getMockedModule(name);

        if(mocked_module){
            return mocked_module;
        }

        var exports = seen[name];
        var oldName = name;
        var mockedDeps = helpers.getMockedDependencies(exports, name);
        var hasMockedDependencies = mockedDeps.length;

        if (exports && !hasMockedDependencies) {
            return exports;
        }

        name = helpers.handleDefaultFilename(name, referrerName);

        exports = seen[name] = seen[oldName] = {};

        var reified = helpers.getDependenciesValues(exports, name);

        // test "instances of objects: should export method" exposes global variable
        // "this" sent as concontext in callback.apply is the cause of the pollution of globalnamespace
        // TODO fix the pollution or find a way to restrict that kind of methods by throwing errors
        registry[name].callback.apply(this, reified);

        return exports;

    };
    helpers.isArray = function(myArray) {
        return Object.prototype.toString.call(myArray).indexOf('Array') > -1;
    };
    helpers.isFunction = function(func){
      return Object.prototype.toString.call(func) === '[object Function]';
    };
    helpers.isObject = function(obj){
      return Object.prototype.toString.call(obj) === '[object Object]';
    };
    helpers.isString = function(string){
      return Object.prototype.toString.call(string) === '[object String]';
    };
    helpers.random = function() {

        var random = Math.round(Math.random() * 100, 2);

        return random;

    };
    helpers.resolve = function(child, name) {

        if (child.charAt(0) === '.') {

            if (!_config.relativePaths) {

                helpers.throwError('Relative paths are not allowed: ' + child + ' required by ' + name);

            }

        } else {

            return child;

        }

        var parts = child.split('/');
        var parentBase = name.split('/').slice(0, -1);

        for (var i = 0, l = parts.length; i < l; i++) {

            var part = parts[i];

            if (part === '..') {

                parentBase.pop();

            } else if (part === '.') {

                continue;

            } else {

                parentBase.push(part);

            }

        }

        return parentBase.join('/');

    };
    helpers.shouldIgnorePrefix = function(path, ignorePatterns) {

        var shouldIgnore = false;

        for (var i = 0; i < ignorePatterns.length; i++) {

            var containsInvalidChars = /[^\w\d/_-]/.test(ignorePatterns[i]);

            if (containsInvalidChars) {

                helpers.throwError('Ignored prefixes contains invalid chars');

            }

            var ignorePattern = new RegExp('^' + ignorePatterns[i] + '(\/)?(\w+)*');

            if (shouldIgnore === false && ignorePattern.test(path)) {

                shouldIgnore = true;

            }

        }

        return shouldIgnore;

    };
    helpers.throwError = function(msg) {

        throw new Error(msg);

    };

    // MODULES

    var define = function(name, deps, callback) {

        var value = {};

        if (!callback) {

            value.deps = [];
            value.callback = deps;

        } else {

            value.deps = deps;
            value.callback = callback;

        }

        if(registry[name]){
            helpers.throwError('duplicatedModule ' + name);
        }

        registry[name] = value;

    };
    var require = function(name, submodule) {

        var module = helpers.internalRequire(name, null);
        var sub_module = module[submodule || 'default'];
        var result = helpers.cloneObject(sub_module);

        return result;

    };
    var has_entries = function(){

        var hasModules = false;

        for (var moduleName in registry) {

            hasModules = !!moduleName;
            break;

        }

        return hasModules;

    };
    var list_entries = function(options){

        var names = [];
        var moduleName;
        var regex = options && options.regex ? options.regex : '';
        var shouldAddModule;

        for (moduleName in registry) {

            shouldAddModule = (regex && moduleName.match(regex)) || !regex;

            if(shouldAddModule){

                names.push(moduleName);

            }

        }

        return names;

    };
    var mark_unseen = function(moduleName){
        seen[moduleName] = undefined;
    };
    var config = function(data) {
        if(data === undefined){

            return helpers.cloneObject(_config);

        } else if(helpers.isString(data)) {

            return helpers.cloneObject(_config[data]);

        } else if(helpers.isObject(data)){

            _config = helpers.extend(defaultConfig, data);

        }

    };
    var mock = function(name, deps, callback){

        Mocked.define(name, deps, callback);

    };

    var reset = function(){

        if(Mocked){

            Mocked.reset();

        } else {

            registry = {};

        }

    };

    config({});

    return {
        define: define,
        require: require,
        config: config,
        has_entries: has_entries,
        list_entries: list_entries,
        mark_unseen: mark_unseen,
        mock: mock,
        reset: reset
    };

};
 module.exports = NMD;